# loemu - A user friendly frontend for game emulators
# 
# Copyright (C) 2007 Bernat Pegueroles Forcadell
# 
# Author: Bernat Pegueroles Forcadell <bernat@pegueroles.cat>
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


import os
import string
import re
import sys
import signal
from pickle import load,dump
from subprocess import Popen,PIPE,STDOUT
from ConfigParser import RawConfigParser
from time import sleep

from tempfile import mkstemp
from cStringIO import StringIO

from xml.sax.handler import ContentHandler
from xml.sax import make_parser
from xml.sax.saxutils import escape

from libxml2 import parseFile
from libxslt import parseStylesheetDoc

from Config import Configuration

# XML Gamelist handler
class XMLHandler(ContentHandler):
    def __init__(self,fields):
        ContentHandler.__init__(self)
        self.record = False
        self.initialize=dict()
        for field in fields:
            self.initialize[field]=""
        self.gamelist=list()
        self.size=0

    def startElement(self, name, attrs):
        if name == "gamelist":
            if "size" in attrs.getNames():
                self.size=float(attrs.get("size"))
        if name == "game":
            game=self.initialize.copy()
            for value in attrs.getNames():
                game[value]=attrs.get(value)
            self.gamelist.append(game)
        
    def get_list(self):
        return self.gamelist

    # returns the part of the gamelist loaded [0,1]
    def get_fraction(self):
        if not self.size:
            return False
        return float(len(self.gamelist) / self.size)

# Gamelist loader
class Loader:
    def __init__(self):
        self.xmlhandler=False
        return

    def load(self):
        self.config.load()
        file = self.config.get("General","gamelist")
        if not os.access(file,os.R_OK):
             return _("Can not access to %s file") % file
        format = self.config.get("General","gamelist_format")
        if format == "xml":
            self.xmlhandler = XMLHandler(self.config.fields)
            saxparser = make_parser()
            saxparser.setContentHandler(self.xmlhandler)
            saxparser.parse(open(file))
            self.list = self.xmlhandler.get_list()
        elif format == "python":
            self.list = load(open(file))
        else:
            return _("Unknown gamelist format") + " ['xml','python']: " + format

        return None

    # returns the part of the gamelist loaded [0,1]
    def get_loader_fraction(self):
        if not self.xmlhandler:
            return False
        return self.xmlhandler.get_fraction()

# Gamelist Manager
class Manager(Loader):
    def __init__(self):
        Loader.__init__(self)
        # configuration values
        self.config = Configuration()
        self.load_configuration()
        # dictionary of current patterns of categories that filter the gamelist
        self.pattern=dict()
        # list off all games of all emulators
        self.list=list()
        # current subset of gamelist
        self.match=list()

    def load_configuration(self):
        self.config.load()

    # return a dictionary of all the possible values of categories of the filters
    def get_filter_lists(self):
        lists=dict()
        self.filter_fraction=0
        states=self.config.get("General","states")
        filters=self.config.get("General","filters")
        for name in filters:
            lists[name]=list()
        for game in self.list:
            self.filter_fraction=self.filter_fraction+1
            if game["state"] in states:
                for name in filters:
                    if game[name] not in lists[name]:
                        lists[name].append(game[name])
        for name in filters:
            lists[name].sort()
            lists[name].insert(0,_("All"))
        return lists

    # returns the part of the filter list loaded [0,1]
    def get_filter_fraction(self):
        return float(float(self.filter_fraction)/len(self.list))

    # set the patterns of categories to execute a new search
    def set_selection(self,category,selection):
        self.pattern=dict()
        if _('All') not in selection:
            self.pattern[category]=("^"+string.join(selection,"$|^")+"$").replace('*','\*')
        self.pattern['emulator']=self.emu
        return

    # fix emulator category in pattern dictionary
    def set_emulator(self,emulator):
        self.emu="^"+emulator+"$"
        self.pattern['emulator']=self.emu
        return

    # filter the gamelist with the dictionary of patterns (or patterns passed)
    def filter(self,pattern=None):
        if pattern:
            self.pattern=pattern
            self.pattern['emulator']=self.emu
        self.match=self.list
        if 'state' not in self.pattern and self.config.has_option('General','states'):
            states=self.config.get('General','states')
            self.pattern['state']=("^"+string.join(states,"$|^")+"$").replace('*','\*')
        for name,value in self.pattern.iteritems():
            if value != "" and value != "^"+_('All')+"$" :
                self.match=self.match_pattern(self.match,name,value)
        return self.match

    # filter the subset of gamelist 'match_list' 
    # with the pattern 'pattern' of category 'name'
    def match_pattern(self,match_list,name,pattern):
        rx=re.compile(pattern,re.I)
        match=list()
        for game in match_list:
            if rx.search(game[name]):
                match.append(game)
        return match

    # return game of current subset of gamelist
    def get(self,index):
        return self.match[index]

    # execute selected game
    def execute(self,index):
        self.load_configuration()
        id=self.match[index]["emulator"]
        emu=self.config.emu(id)

        rx=re.compile("@\((?P<replace>[^)]+)\)")
        args=list() 
        for arg in emu.get("General","start").split(" "):
            if arg != '':
                mo = re.search(rx,arg)
                if mo:
                    args.append(re.sub(rx,self.match[index][mo.group('replace')],arg))
                else:
                    args.append(arg)

        p=Popen(args,close_fds=True,stderr=STDOUT,stdout=PIPE)
        return p

# Gamelist Builder
class Builder(Loader):
    def __init__(self):
        Loader.__init__(self)
        # configuration values
        self.config = Configuration()
        self.config.load()
        if not os.access(os.path.dirname(self.config.get("General","gamelist")),os.W_OK):
            self.log("99",_("Can not write gamelist file: %s") % self.config.get("General","gamelist"))
            sys.exit(1)
            return
        self.out= sys.stdout
        # set signal handler to handle extern kill
        signal.signal(signal.SIGTERM,self.signal_handler)

    # build the gamelist of all emulators
    def build(self):
        self.list=list()
        for id in self.config.emulators():
            self.build_id(id)
            self.list = self.list + self.current.values()
        self.write()

    # build the gamelist with id 'id'
    def build_emulator(self,id):
        if not os.access(self.config.get("General","gamelist"),os.R_OK):
             self.list=list()
        else:
            self.log("10",_("Loading gamelist ..."))
            self.log("11",self.config.get("General","gamelist"))
            self.log("00"," ")
            self.log("20","-1")
            err=self.load()
            if err:
                self.log("99",err)
        self.id=id
        # remove the emulator that has to be rebuilded
        self.list=filter(self.remove_id,self.list)
        self.build_id(id)
        self.list = self.list + self.current.values()
        self.write()

    def remove_id(self,game):
        if game['emulator'] == self.id:
            return False
        return True

    # build the subset of gamelist of emulator 'id'
    def build_id(self,id):
        self.current=dict()
        emu=self.config.emu(id)

        if not os.access(emu.get("General","binary"),os.X_OK):
            for path in os.environ['PATH'].split(":"):
                 if os.access(path + "/" + emu.get("General","binary"),os.X_OK):
                    break
            else:
                return

        for section in emu.get("General","sources"):
            type=emu.get(section,"type")
            source=emu.get(section,type)
            format = emu.get(section,"format")

            if type=="archive":
                src = None
                if os.access(source,os.R_OK):
                    src = open(source)
                self.log("10",_("Loading %s file ...") % format)
                self.log("11",source)
            elif type=="command":
                p=Popen(source.split(" "),close_fds=True,stdout=PIPE,stderr=PIPE)
                src = p.stdout
                self.log("10",_("Loading %s output ...") % format)
                self.log("11",source)
            else:
                self.log("99",_("Type of source unknown") + " ['archive','command']: " + type)
                sys.exit(1)
                return

            if format == "ini-sections":
                self.load_ini_sections_source(src,emu,section)
            elif format == "ini-options":
                self.load_ini_options_source(src,emu,section)
            elif format == "xml":
                self.load_xml_source(src,emu,section)
            elif format == "plain":
                self.load_plain_source(src,emu,section)
            else:
                self.log("99",_("Format of source unknown") + " ['xml','ini-sections','ini-options','plain']: " + format)
                sys.exit(1)
                return

        # transform the values returned to the standarized values of all emulators
        #  (example in state category: correct -> good, incorrect -> bad, ...)
        transform_values=dict()
        for field in self.config.fields:
            if field in emu.options('General'):
                transform_values[field]=emu.get('General',field)
        for game in self.current:
            self.current[game]["emulator"]=id
            for key,value in self.current[game].iteritems():
                if key in transform_values:
                    self.current[game][key]=transform_values[key].get(value,value)
        return

    def load_ini_sections_source(self,src,emu,section):
        if src == None:
            return
        self.log("20","-1")
        self.log("00",_("Initializing ..."))
        ilabel=emu.get("General","identity")
        sections=emu.get(section,"sections")

        parser= RawConfigParser()
        parser.readfp(src)
        for title,key in sections.iteritems():
            self.log("00",_("Parsing %s section ...") % title)
            if parser.has_section(title):
                for ident,value in parser.items(title):
                    if not self.current.has_key(ident):
                        self.current[ident]=dict()
                        for field in self.config.fields:
                            self.current[ident][field]=""
                        self.current[ident][ilabel]=ident
                    if self.current[ident][key]=="":
                        self.current[ident][key]=value
        return 

    def load_ini_options_source(self,src,emu,section):
        if src == None:
            return
        self.log("20","-1")
        self.log("00",_("Initializing ..."))
        ilabel=emu.get("General","identity")
        sections=emu.get(section,"sections")

        parser= RawConfigParser()
        parser.readfp(src)
        for ident in parser.sections():
            self.log("00",_("Parsing %s section ...") % ident)
            for title,value in parser.items(ident):
                if options.has_key(title):
                    if not self.current.has_key(ident):
                        self.current[ident]=dict()
                        for field in self.config.fields:
                            self.current[ident][field]=""
                        self.current[ident][ilabel]=ident
                    if self.current[ident][options[title]]=="":
                        self.current[ident][options[title]]=value
        return

    def load_xml_source(self,src,emu,section):
        if src == None:
            return
        self.log("00",_("Initializing ..."))
        self.log("20","0.125")
        if os.access(emu.get(section,"xsl"),os.R_OK):
            xslfile=emu.get(section,"xsl")
        elif os.access(self.config.config_dir + emu.get(section,"xsl"),os.R_OK):
            xslfile=self.config.config_dir + emu.get(section,"xsl")
        else:
            return
        tempfile=mkstemp()
        os.write(tempfile[0],src.read())

        self.log("00",_("Parsing XSL file (%s)...") % xslfile)
        styledoc = parseFile(xslfile)
        style = parseStylesheetDoc(styledoc)
        self.log("00",_("Parsing xml output ..."))
        self.log("20","0.250")
        doc = parseFile(tempfile[1])
        self.log("00",_("Applying XSL (%s) ...") % xslfile)
        self.log("20","0.375")
        result = style.applyStylesheet(doc, None)

        self.log("00",_("Saving result ..."))
        self.log("20","0.500")
        str= style.saveResultToString(result)
        style.freeStylesheet()
        doc.freeDoc()
        result.freeDoc()

        self.log("00",_("Initializing result ..."))
        self.log("20","0.625")
        sio=StringIO(str)
        handler = XMLHandler(self.config.fields)
        saxparser = make_parser()
        saxparser.setContentHandler(handler)
        self.log("00",_("Parsing result ..."))
        self.log("20","0.750")
        saxparser.parse(sio)

        self.log("00",_("Finalizing ..."))
        self.log("20","0.875")
        ilabel=emu.get("General","identity")
        for game in handler.get_list():
            ident=game[ilabel].strip()
            if not self.current.has_key(ident):
                self.current[ident]=dict()
                for field in self.config.fields:
                    self.current[ident][field]=""
            for key,value in game.iteritems():
                if self.current[ident][key] == "":
                    self.current[ident][key]=value
        self.log("20","1.0")
        return True

    def load_plain_source(self,src,emu,section):
        if src == None:
            return
        self.log("20","0.0")
        self.log("00",_("Initializing ..."))
        ilabel=emu.get("General","identity")
        rx=re.compile(emu.get(section,"pattern"),re.I)
        current_game=0.0
        all_games=float(len(self.current)) or -1
        for line in src:
            match=rx.search(line)
            if match != None:
                current_game=current_game+1.0
                ident=match.group(ilabel)
                self.log("00",_("Parsing %s ...") % ident)
                self.log("20",str(float(current_game/all_games)))
                if not self.current.has_key(ident):
                    self.current[ident]=dict()
                    for field in self.config.fields:
                        self.current[ident][field]=""
                    self.current[ident][ilabel]=ident
                for key,value in match.groupdict().iteritems():
                    if key != ilabel and self.current[ident][key]=="":
                        self.current[ident][key]=value
        return

    def write(self):
        self.log("90",_("Writing ..."))
        self.log("11",self.config.get("General","gamelist"))
        self.log("00"," ")
        signal.signal(signal.SIGTERM,signal.SIG_IGN)
        format = self.config.get("General","gamelist_format")
        if  format == "xml":
            self.write_xml(self.config.get("General","gamelist"))
        elif format == "python":
            dump(self.list,open(self.config.get("General","gamelist"),"w"))
        else:
            self.log("99",_("Unknown gamelist format") + " ['xml','python']: " + format)
            sys.exit(1)

    def write_xml(self,file):
        self.log("20","-1")
        fields=self.config.fields
        fd=open(file,'w')
        fd.write('<?xml version="1.0" ?>\n')
        fd.write('<gamelist version="1.1" size="%d">\n' % len(self.list))
        current_game=0.0
        all_games=float(len(self.list)) or -1
        for game in self.list:
            current_game=current_game+1.0
            self.log("20",str(float(current_game/all_games)))
            raw = '  <game'
            for field in fields:
                raw += ' %s="%s"' % (field, escape(game[field].encode('utf-8')))
            raw += ' />\n'
            fd.write(raw)
        fd.write('</gamelist>\n')
        fd.close()
        return

    def log(self,level,msg):
        self.out.write(level + " " + escape(msg) + "\n")
        self.out.flush()
        return

    def signal_handler(self,signum,frame):
        sys.exit(0)
        return
